package es.uca.xmleye.gui;

/*
 Copyright (C) 2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import javax.swing.*;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.text.html.*;
import javax.swing.tree.TreePath;

import java.util.Locale;
import java.util.ResourceBundle;
import java.util.regex.*;

import es.uca.xmleye.ParsingException;
import es.uca.xmleye.SearchException;
import es.uca.xmleye.gui.html.*;
import es.uca.xmleye.gui.trees.*;
import es.uca.xmleye.presentation.DocumentPresentationModel;

/**
 * Vista por defecto de un documento.
 */
class DocumentView extends JPanel {
	/**
	 * 
	 */
	private static final long serialVersionUID = -2918959531266698618L;
	private static final String HOJA_CSS = "css/default.css";
	private static final ResourceBundle errRes;

	static {
		errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
				Locale.getDefault());
	}

	private JScrollPane panelIzq;
	private DOMTree arbolDoc;
	private JScrollPane panelDer;
	private XHTMLViewerPane panelHTML;

	public DocumentView() throws ParsingException {
		super(new BorderLayout());

		// LEFT PANEL - DOCUMENT TREE
		arbolDoc = new DOMTree(new DataDrivenTreeModel());
		arbolDoc.setCellRenderer(new DataDrivenTreeCellRenderer());
		panelIzq = new JScrollPane(arbolDoc);
		panelIzq.setMinimumSize(new Dimension(200, 500));

		// RIGHT PANEL - NODE VIEW (i.e.: XHTML)
		panelHTML = new XHTMLViewerPane("text/html", "");
		panelHTML.setEditable(false);

		// panelHTML.addHyperlinkListener(new DebugLinkListener());
		panelHTML.addHyperlinkListener(new ObservadorLinkInterno());
		panelHTML.addHyperlinkListener(new AnchorLinkListener());
		panelHTML.addHyperlinkListener(new ExternalBrowserLinkListener());
		panelHTML.addHyperlinkListener(new HoverIconLinkListener(this,
				panelHTML));

		// Añadimos reglas CSS
		final HTMLEditorKit htmlKit = (HTMLEditorKit) panelHTML.getEditorKit();
		try {
			final Reader cssReader = new FileReader(HOJA_CSS);
			htmlKit.getStyleSheet().loadRules(cssReader, null);
		} catch (FileNotFoundException fEx) {
			MessageManagerSingleton.getInstancia().mostrarError(
					errRes.getString("err_cssnotfound"));
		} catch (IOException ioEx) {
			MessageManagerSingleton.getInstancia().mostrarError(
					errRes.getString("err_cssioex"));
		}

		panelDer = new JScrollPane();
		panelHTML.instalar(panelDer);

		// PANEL DIVISOR
		JSplitPane panelSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
				panelIzq, panelDer);
		panelSplit.setOneTouchExpandable(true);
		panelSplit.setBorder(null);

		add(panelSplit, BorderLayout.CENTER);

		// PROPIEDADES
		setPreferredSize(new Dimension(WndMain.ANCHO_MIN, WndMain.ALTO_MIN));
		setBorder(null);
	}

	/**
	 * Añadimos a <code>listener</code> para su notificación ante activaciones
	 * de un enlace determinado.
	 */
	public void addHyperlinkListener(HyperlinkListener listener) {
		panelHTML.addHyperlinkListener(listener);
	}

	/**
	 * Retiramos a <code>listener</code> para su notificación ante
	 * activaciones de un enlace determinado.
	 */
	public void removeHyperlinkListener(HyperlinkListener listener) {
		panelHTML.removeHyperlinkListener(listener);
	}

	/**
	 * Añadimos a <code>listener</code> para su notificación ante selecciones
	 * de un elemento determinado.
	 */
	public void addTreeSelectionListener(TreeSelectionListener listener) {
		arbolDoc.addTreeSelectionListener(listener);
	}

	/**
	 * Retiramos a <code>listener</code> para su notificación ante selecciones
	 * de un elemento determinado.
	 */
	public void removeTreeSelectionListener(TreeSelectionListener listener) {
		arbolDoc.removeTreeSelectionListener(listener);
	}

	/**
	 * Cambia el nodo seleccionado al padre del nodo actual, si lo tiene. De lo
	 * contrario, no hace nada.
	 */
	public void irPadre() {
		arbolDoc.selectParent();
	}

	/**
	 * Cambia el nodo seleccionado al primer hijo del nodo actual, si lo tiene.
	 * De lo contrario, no hace nada.
	 */
	public void irPrimerHijo() {
		arbolDoc.selectFirstChild();
	}

	/**
	 * Cambia el nodo seleccionado al siguiente hermano del nodo actual, si lo
	 * tiene. De lo contrario, no hace nada.
	 */
	public void irSigHermano() {
		arbolDoc.selectNextSibling();
	}

	/**
	 * Cambia el nodo seleccionado al anterior hermano del nodo actual, si lo
	 * tiene. De lo contrario, no hace nada.
	 */
	public void irAntHermano() {
		arbolDoc.selectPreviousSibling();
	}

	/**
	 * Expande al máximo el árbol. No modifica la selección.
	 */
	public void expandirTodo() {
		for (int i = 0; i < arbolDoc.getRowCount(); ++i) {
			arbolDoc.expandRow(i);
		}
	}

	/**
	 * Contrae totalmente el árbol, y selecciona la raíz.
	 */
	public void contraerTodo() {
		for (int i = arbolDoc.getRowCount() - 1; i >= 0; --i) {
			arbolDoc.collapseRow(i);
		}
	}

	/**
	 * Sincronizamos con un modelo de documento.
	 */
	public void sincronizar(DocumentPresentationModel mod) {
		// Al cambiar de modelo del árbol, perderemos la selección debido a los
		// observadores que normalmente monitorizan los cambios en la selección.
		// Hemos de guardarla en una variable temporal.
		TreePath seleccion = mod.getSeleccion();
		if (arbolDoc.getModel() != mod.getModArbol())
			arbolDoc.setModel(mod.getModArbol());

		if (seleccion != null
				&& !seleccion.equals(arbolDoc.getSelectionModel().getSelectionPath()))
			arbolDoc.getSelectionModel().setSelectionPath(seleccion);

		arbolDoc.scrollPathToVisible(seleccion);
		panelHTML.setText(mod.getFuenteXHTML());
	}

	/** CLASES INTERNAS * */
	private class ObservadorLinkInterno implements HyperlinkListener {
		public void hyperlinkUpdate(HyperlinkEvent e) {
			if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED
					&& e.getDescription() != null && e.getURL() == null) {
				try {
					final Matcher m = Pattern.compile("#xpointer\\((.*)\\)")
							.matcher(e.getDescription());

					if (m.matches()) {
						arbolDoc.selectNode(m.group(1));
					}
				} catch (SearchException ex) {
					MessageManagerSingleton.getInstancia().mostrarError(
							errRes.getString("err_path_invalid") + "\n"
									+ ex.getLocalizedMessage());
				}
			}
		}
	}
}
